// Copyright (c) Graham Relf, UK, 2021
// www.grelf.net
// A demonstration of using the 2D graphics context of HTML5 to create animated
// 3D worlds from plain JavaScript without any other libraries or frameworks.
'use strict';

// Dominoes + Dice

// Create lying horizontally, pips up. Colours are {r:r, g:g, b:b}
function Domino (xCentre, yCentre, zCentre, length, no1, no2, bodyColour, pipColour)
{ var hwd = length / 2, hdp = length / 4, hht = length / 20;
  var colours = [];
  for (var i = 0; i < 6; i++) colours.push (bodyColour);
  Cuboid.call (this, xCentre, yCentre, zCentre, hwd, hdp, hht, colours);
  this.setStroking (true);
  this.setStroke (100, 100, 100);
  this.pipSize = length / 30;
  this.pipColour = pipColour;
  this.qwd = hwd / 2;
  this.pips = [];
  this.x1 = xCentre - this.qwd;
  this.x2 = xCentre + this.qwd;
  this.y = yCentre;
  this.zTop = zCentre + hht;
  this.addPips (new Point3D (this.x1, this.y, this.zTop), no1);
  this.addPips (new Point3D (this.x2, this.y, this.zTop), no2);
}

Domino.prototype = Object.create (Cuboid.prototype);
Domino.prototype.constructor = Domino;

// Pip positions relative to centre of half-domino
Domino.prototype.PATTERNS =
[[], // 0
 [{x:0, y:0}], // 1
 [{x:-0.4, y:-0.4}, {x:0.4, y:0.4}], // 2
 [{x:-0.4, y:-0.4}, {x:0, y:0}, {x:0.4, y:0.4}], // 3
 [{x:-0.4, y:-0.4}, {x:0.4, y:0.4}, {x:-0.4, y:0.4}, {x:0.4, y:-0.4}], // 4
 [{x:-0.4, y:-0.4}, {x:0.4, y:0.4}, {x:0, y:0}, {x:-0.4, y:0.4}, {x:0.4, y:-0.4}], // 5
 [{x:-0.4, y:-0.4}, {x:0.4, y:0.4}, {x:0, y:-0.4}, {x:0, y:0.4},
  {x:-0.4, y:0.4}, {x:0.4, y:-0.4}] // 6
];

Domino.prototype.addPips = function (centre, no)
{ if (no < 0 || no > 6) error ("Invalid domino number " + no);
  var pn = this.PATTERNS [no];
  for (var j = 0; j < pn.length; j++)
  { var pnj = pn [j];
    var pip = createHexagon (centre.x + this.qwd * pnj.x, centre.y + this.qwd * pnj.y, this.pipSize);
    pip.setColour (this.pipColour.r, this.pipColour.g, this.pipColour.b);
    pip.translate (0, 0, centre.z);
    this.pips.push (pip);
    this.faces [0].addDecoration (pip);
  }
};

// Can change pips if domino has been translated but not if it has rotated at all
Domino.prototype.changePips = function (no1, no2)
{ this.faces [0].decorations = [];
  this.pips = [];
  this.addPips (new Point3D (this.x1, this.y, this.zTop), no1);
  this.addPips (new Point3D (this.x2, this.y, this.zTop), no2);
};

Domino.prototype.translate = function (dx, dy, dz)
{ this.x1 += dx;
  this.x2 += dx;
  this.y += dy;
  this.zTop += dz;
  Shape.prototype.translate.call (this, dx, dy, dz);
};

function Die (xCentre, yCentre, zCentre, width, bodyColour, pipColour)
{ this.hwd = width / 2;
  var colours = [];
  for (var i = 0; i < 6; i++) colours.push (bodyColour);
  Cuboid.call (this, xCentre, yCentre, zCentre, this.hwd, this.hwd, this.hwd, colours);
  this.radius = Math.sqrt (3 * this.hwd * this.hwd);
  this.setStroking (true);
  this.setStroke (100, 100, 100);
  this.pipSize = width / 12;
  this.pipColour = pipColour;
  var nos = [6, 1, 5, 2, 4, 3];
  for (i = 0; i < 6; i++) this.addPips (i, nos [i]);
  this.landed = false;
  this.UNIT_VERTICAL = new Vector3D (0, 0, 1);
}

Die.prototype = Object.create (Cuboid.prototype);
Die.prototype.constructor = Die;

Die.prototype.addPips = function (i, no)
{ var pn = Domino.prototype.PATTERNS [no];
  var face = this.faces [i];
  if (i === 2 || i=== 3) this.rotateX (90);
  else if (i === 4 || i === 5) this.rotateY (90);
  for (var j = 0; j < pn.length; j++)
  { var pnj = pn [j];
    var pip = createHexagon (this.hwd * pnj.x, this.hwd * pnj.y, this.pipSize);
    pip.translate (face.centre.x, face.centre.y, face.centre.z);
    pip.setColour (this.pipColour.r, this.pipColour.g, this.pipColour.b);
    face.addDecoration (pip);
  }
  if (i === 2 || i=== 3) this.rotateX (-90);
  else if (i === 4 || i === 5) this.rotateY (-90);
};

Die.prototype.toss = function (dx, dy, dz, ax, ay, az)
{ // Makes movable => use update()
  this.movable = false;
  this.landed = true; // Prevent update() until ready
  this.velocity = new Vector3D (dx, dy, dz); // Upwards
  this.setAngular (ax, ay, az);
  this.landed = false;
};

Die.prototype.update = function ()
{ if (!this.landed)
  { this.rotateX (this.angular.dx);
    this.rotateY (this.angular.dy);
    this.rotateZ (this.angular.dz);
    // Reflect off edge of world:
    if (this.centre.x > world.range
     || this.centre.x < -world.range) this.velocity.dx = -this.velocity.dx;
    if (this.centre.y > world.range
     || this.centre.y < -world.range) this.velocity.dy = -this.velocity.dy;
    this.velocity.dz -= world.g;
    var newZ = this.centre.z + this.velocity.dz;
    if (newZ > this.halfheight)
    { this.translate (this.velocity.dx, this.velocity.dy, this.velocity.dz); }
    else
    { this.landed = true;
      var frac = (this.centre.z - this.halfheight) / newZ;
      this.translate (frac * this.velocity.dx, frac * this.velocity.dy, frac * this.velocity.dz);
      // Find highest face:
      var maxF = this.faces [0], maxFz = maxF.centre.z, maxI = 0;
      for (var i = 1; i < this.faces.length; i++)
      { var fz = this.faces [i].centre.z;
        if (fz > maxFz)
        { maxFz = fz;
          maxF = this.faces [i];
          maxI = i;
      } }
      var isOdd = (maxI % 2 === 1);
      maxF.setUnitNormal ();
      var nv = maxF.unitNormal;
      var axDegs = RAD2DEG * Math.atan2 (nv.dy, nv.dz);
      if (isOdd) axDegs -= 180;
      this.rotateX (-axDegs);
      maxF.setUnitNormal ();
      nv = maxF.unitNormal;
      var ayDegs = RAD2DEG * Math.atan2 (nv.dx, nv.dz);
      if (isOdd) ayDegs -= 180;
      this.rotateY (-ayDegs);
  } }
};