// Copyright (c) Graham Relf, UK, 2020
// www.grelf.net
'use strict';

// iVerts is array of indices into shape.vertices array
function Face (parentShape, centre, vertices, colour, iVerts)
{ this.parent = parentShape;
  this.centre = centre; // Point3D
  this.vertices = vertices; // Array of Point3D
  this.colour = colour; // fill { r, g, b }
  this.iVerts = iVerts;
  this.stroke = {r:0, g:0, b:0}; // Black default
  this.stroking = true;
  this.im = null;
  this.decorations = []; // New in faceC
}

Face.prototype.addDecoration = function (shape) { this.decorations.push (shape); };

Face.prototype.recentre = function () // Needed after scaling: see Shape
{ var sumX = 0, sumY = 0, sumZ = 0, ivLen = this.vertices.length;
  for (var iv = 0; iv < ivLen; iv++)
  { this.vertices [iv] = this.parent.vertices [this.iVerts [iv]];
    var v = this.vertices [iv];
    sumX += v.x; sumY += v.y; sumZ += v.z;
  }
  this.centre = new Point3D (sumX / ivLen, sumY / ivLen, sumZ / ivLen); // mean values
};

Face.prototype.getNormalVector = function ()
{ var v1 = this.vertices [0].difference (this.vertices [1]);
  var v2 = this.vertices [1].difference (this.vertices [2]);
  return v1.cross (v2);
};

Face.prototype.setUnitNormal = function ()
{ var n = this.getNormalVector ();
  n.scale (1 / n.magnitude ());
  this.unitNormal = n;
};

Face.prototype.getPoint3DDistance = function (pt3d)
{ if (!this.unitNormal) this.setUnitNormal ();
  var q = pt3d.difference (this.centre);
  return q.dot (this.unitNormal);
};

Face.prototype.setSolid = function ()
{ if (this.vertices.length < 2) { error ('Face has too few vertices for setSolid()'); return; }
  var v0 = this.vertices [0];
  var minX = v0.x, minY = v0.y, minZ = v0.z;
  var maxX = minX, maxY = minY, maxZ = minZ;
  for (var i = 1; i < this.vertices.length; i++)
  { var vi = this.vertices [i];
    if (vi.x < minX) minX = vi.x; else if (vi.x > maxX) maxX = vi.x;
    if (vi.y < minY) minY = vi.y; else if (vi.y > maxY) maxY = vi.y;
    if (vi.z < minZ) minZ = vi.z; else if (vi.z > maxZ) maxZ = vi.z;
  }
  if (minZ === maxZ) { error ('Face must be vertical for setSolid()'); return; }
  if (minX !== maxX && minY !== maxY)
  { error ('Face must be normal to x- or y-axis for setSolid()'); return; }
  if (minX !== maxX)
  { world.solidWalls.push
    ( { face: this, minX: minX, maxX: maxX, maxZ: maxZ,
        tooClose: function (pt3d)
        { if (pt3d.x >= this.minX && pt3d.x <= this.maxX && pt3d.z <= this.maxZ
           && Math.abs (this.face.getPoint3DDistance (pt3d)) < world.CLOSEST) return true;
          return false;
      } }
    );
  }
  else if (minY !== maxY)
  { world.solidWalls.push
    ( { face: this, minY: minY, maxY: maxY, maxZ: maxZ,
        tooClose: function (pt3d)
        { if (pt3d.y >= this.minY && pt3d.y <= this.maxY && pt3d.z <= this.maxZ
           && Math.abs (this.face.getPoint3DDistance (pt3d)) < world.CLOSEST) return true;
          return false;
      } }
    );
  }
  else error ("setSolid failed");
};

Face.prototype.setAlpha = function (a) { this.alpha = a; };// 0..1
Face.prototype.setColour = function (r, g, b) { this.colour = {r:r, g:g, b:b}; };
Face.prototype.setStroke = function (r, g, b) { this.stroke = {r:r, g:g, b:b}; };
Face.prototype.setStroking = function (trueOrFalse) { this.stroking = trueOrFalse; };
Face.prototype.setImage = function (imageName) { this.im = world.images [imageName]; };

Face.prototype.view = function (me, fogNo)
{ if (null !== this.im) this.mapImage (me, fogNo);
  else
  { var g2 = world.g2;
    g2.beginPath ();
    var screenPts = [];
    var sxy = me.getScreenXYD (this.vertices [0]);
    screenPts.push (sxy);
    g2.moveTo (sxy.x, sxy.y);
    for (var j = 1; j < this.vertices.length; j++)
    { sxy = me.getScreenXYD (this.vertices [j]);
      screenPts.push (sxy);
      g2.lineTo (sxy.x, sxy.y);
    }
    g2.closePath ();
    g2.fillStyle = fog (this.colour.r, this.colour.g, this.colour.b, fogNo);
    g2.globalAlpha = this.alpha;
    g2.fill ();
    g2.globalAlpha = 1;
    if (this.stroking)
    { g2.strokeStyle = fog (this.stroke.r, this.stroke.g, this.stroke.b, fogNo);
      g2.stroke ();
    }
    for (j = 0; j < this.decorations.length; j++) this.decorations [j].view (me, fogNo);
    var target;
    if (this.isTarget)
    { target = this;
      this.minX = null; this.maxX = null; this.minY = null; this.maxY = null;
    }
    else if (this.parent.isTarget) { target = this.parent; }
    if (undefined !== target)
    { var b = getBoundsXY (screenPts);
      if (null !== b)
      { if (null === target.minX) target.minX = b.minX; 
        else target.minX = Math.min (target.minX, b.minX); 
        if (null === target.maxX) target.maxX = b.maxX; 
        else target.maxX = Math.max (target.maxX, b.maxX); 
        if (null === target.minY) target.minY = b.minY; 
        else target.minY = Math.min (target.minY, b.minY); 
        if (null === target.maxY) target.maxY = b.maxY; 
        else target.maxY = Math.max (target.maxY, b.maxY);
  } } }
};

/* This is not the usual pixel-by-pixel inverse transform algorithm. It should be faster than that 
because it writes whole rows to the screen data */
Face.prototype.mapImage = function (me, fogNo)
{ if (!this.im.forFaces) { error ("Face.mapImage: image not set for use on face"); return; }
  if (this.vertices.length !== 4) { error ("Face.mapImage: face must have 4 vertices"); return; } // FOR NOW
  var sxy = [];
  for (var i = 0; i < this.vertices.length; i++)
  { sxy.push (me.getScreenXYD (this.vertices [i])); }
  // Assume vertices clockwise from top left of image
  var bounds = getBoundsXY (sxy);
  if (null === bounds) return; // All off screen
  var bxy = []; // Coordinates within bounds
  for (i = 0; i < 4; i++)
  { var si = sxy [i];
    bxy.push ({x:Math.round (si.x - bounds.minX), y:Math.round (si.y - bounds.minY)});
  }
  
  // First version used a mask image but 2nd path version is probably faster:
//  var maskData = createMask (bounds, bxy);
//world.g2.putImageData (maskData, 0, 0);// test
world.g2.beginPath ();
world.g2.moveTo (bxy [0].x, bxy [0].y);
for (var i = 1; i < bxy.length; i++) world.g2.lineTo (bxy [i].x, bxy [i].y);
world.g2.closePath ();

  // Get screenData but cropped to be only on screen:
  var crMinX = bounds.minX + bounds.marginL;
  var crMinY = bounds.minY + bounds.marginT;
  var crWd = bounds.wd - bounds.marginL - bounds.marginR;
  var crHt = bounds.ht - bounds.marginT - bounds.marginB;
  if (crHt <= 0 || crWd <= 0) return; // Just off screen
  var screenData = world.g2.getImageData (crMinX, crMinY, crWd, crHt);
//world.g2.putImageData (screenData, 0, 0);// test

  // Get data of source image:
  var imData = this.im.foggy [Math.round (fogNo * 7)].data;
//world.g2.putImageData (imData, 0, 0);// test

  var iT = bounds.iTop;
  var iNext = iT + 1; if (iNext > 3) iNext = 0;
  var iPrev = iT - 1; if (iPrev < 0) iPrev = 3;
  var iLeft0, iLeft1, iRight0, iRight1;
  if (bxy [iPrev].y === bxy [iT].y) { iLeft0 = iPrev; iRight0 = iT; } // Flat top case 1
  else if (bxy [iNext].y === bxy [iT].y) { iLeft0 = iT; iRight0 = iNext; } // Flat top case 2
  else iRight0 = iLeft0 = iT; // Single pt at top
  iLeft1 = iLeft0 - 1; if (iLeft1 < 0) iLeft1 = 3;
  iRight1 = iRight0 + 1; if (iRight1 > 3) iRight1 = 0;
  var dLeft = distance (bxy [iLeft0], bxy [iLeft1]);
  var dRight = distance (bxy [iRight0], bxy [iRight1]);

  for (var dstY = bxy [iT].y; dstY < bounds.ht - bounds.marginB; dstY++)
  { var dstMinX = 0, dstMaxX = bounds.wd - 1;
//    while (maskData.getPixel (dstMinX, dstY) [0] === 0 && dstMinX < bounds.wd - 1) dstMinX++;
//    while (maskData.getPixel (dstMaxX, dstY) [0] === 0 && dstMaxX > 0) dstMaxX--;
while (!world.g2.isPointInPath (dstMinX, dstY) && dstMinX < bounds.wd - 1) dstMinX++;
while (!world.g2.isPointInPath (dstMaxX, dstY) && dstMaxX > 0) dstMaxX--;
    // We now know the line segment on the mask. Find it on the source image:
    var ratioLeft = distance (bxy [iLeft0], {x:dstMinX, y:dstY}) / dLeft;
    var ratioRight = distance (bxy [iRight0], {x:dstMaxX, y:dstY}) / dRight;
    var imiL0 = this.im.corners [iLeft0], imiR0 = this.im.corners [iRight0];
    var imiL1 = this.im.corners [iLeft1], imiR1 = this.im.corners [iRight1];
    var srcX0 = imiL0.x + ratioLeft * (imiL1.x - imiL0.x);
    var srcY0 = imiL0.y + ratioLeft * (imiL1.y - imiL0.y);
    var srcX1 = imiR0.x + ratioRight * (imiR1.x - imiR0.x);
    var srcY1 = imiR0.y + ratioRight * (imiR1.y - imiR0.y);
    // Plot row:
    var dstYadj = dstY - bounds.marginT;
    var dstLen = dstMaxX - dstMinX;
    var srcDx = (srcX1 - srcX0) / dstLen; // May be negative of course
    var srcDy = (srcY1 - srcY0) / dstLen; // ..
    var srcX = srcX0, srcY = srcY0;
    var rowX0 = Math.max (dstMinX, bounds.marginL);
    var rowX1 = Math.min (dstMaxX, bounds.wd - bounds.marginR);
    var rowLen = (rowX1 - rowX0) * 4;
    if (rowLen > 0)
    { var row = new Uint8ClampedArray (rowLen);
      var iRow = 0;
      for (var dstX = dstMinX; dstX <= dstMaxX && srcX < imData.width && srcY < imData.height; dstX++)
      { if (dstX >= bounds.marginL && dstX < bounds.wd - bounds.marginR)
        { var px = imData.getPixel (srcX, srcY);
          for (var iPx = 0; iPx < 4; iPx++, iRow++) row [iRow] = px [iPx];
        }
        srcX += srcDx;
        srcY += srcDy;
      }
      screenData.setRow (rowX0 - bounds.marginL, dstYadj, row);
    }
    if (dstMinX === bxy [iLeft1].x && dstY === bxy [iLeft1].y)
    { iLeft0 = iLeft1;
      iLeft1--; if (iLeft1 < 0) iLeft1 = 3;
      dLeft = distance (bxy [iLeft0], bxy [iLeft1]);
    }
    if (dstMaxX >= bxy [iRight1].x - 1 && dstY === bxy [iRight1].y)
    { iRight0 = iRight1;
      iRight1++; if (iRight1 > 3) iRight1 = 0;
      dRight = distance (bxy [iRight0], bxy [iRight1]);
  } }
  world.g2.putImageData (screenData, crMinX, crMinY);
};

/*function createMask (bounds, pts)
{ var cnv = document.createElement ('canvas');
  cnv.width = bounds.wd;
  cnv.height = bounds.ht;
  var g2 = cnv.getContext ('2d');
  g2.fillStyle = '#000';
  g2.fillRect (0, 0, bounds.wd, bounds.ht);
  g2.beginPath ();
  g2.moveTo (pts [0].x, pts [0].y);
  for (var i = 1; i < pts.length; i++) g2.lineTo (pts [i].x, pts [i].y);
  g2.closePath ();
  g2.fillStyle = '#fff';
  g2.fill ();
  return g2.getImageData (0, 0, cnv.width, cnv.height);
}*/

/* Returns null if all off screen */
function getBoundsXY (pts) // Array of {x:, y:, ...}
{ var minX = pts [0].x, minY = pts [0].y, maxX = minX, maxY = minY, iTop = 0;
  for (var i = 1; i < pts.length; i++)
  { var p = pts [i];
    if (p.x < minX) minX = p.x;
    else if (p.x > maxX) maxX = p.x;
    if (p.y < minY) { minY = p.y; iTop = i; }
    else if (p.y > maxY) maxY = p.y;
  }
  if (minX >= world.screenWd || maxX < 0 || minY >= world.screenHt || maxY < 0) return null;
  minX = Math.round (minX);
  maxX = Math.floor (maxX);
  minY = Math.round (minY);
  maxY = Math.floor (maxY);
  var marginL = (minX < 0) ? -minX : 0;
  var marginR = (maxX >= world.screenWd) ? maxX - world.screenWd + 1 : 0;
  var marginT = (minY < 0) ? -minY : 0;
  var marginB = (maxY >= world.screenHt - 1) ? maxY - world.screenHt + 1 : 0;
  return ({maxX:maxX, minX:minX, maxY:maxY, minY:minY, iTop: iTop,
    wd:maxX - minX + 1, ht:maxY - minY + 1,
    marginL:marginL, marginR:marginR, marginT:marginT, marginB:marginB});
}

Face.prototype.explode = function ()
{ var c = this.centre;
  for (var i = 0; i < this.vertices.length; i++)
  { var i1 = i + 1; if (i1 >= this.vertices.length) i1 = 0;
    var v0 = this.vertices [i], v1 = this.vertices [i1];
    var frag = new Polygon (c.x, c.y, [c, v0, v1]);
    frag.centre.z = c.z;
    frag.setVelocity (random (-0.5, 0.5), random (-0.5, 0.5), random (0.3, 1)); // All upwards
    frag.setAngular (random (-5, 5), random (-5, 5), random (-5, 5));
    frag.setColour (this.colour.r, this.colour.g, this.colour.b);
    frag.setFreeFall (true);
    world.add (frag);
    if (this.parent.isTarget) world.score += 100; // Not a timed explosion
  }
  world.remove (this.parent);
  for (i = 0; i < world.solidWalls.length; i++)
  { if (world.solidWalls [i].face === this) world.solidWalls.splice (i, 1); }
  animate ();
};