// Copyright (c) Graham Relf, UK, 2020
// www.grelf.net
'use strict';

// An abstract type. Inherited by all other shapes but not to be instantiated itself
function Shape ()
{ this.centre = undefined;
  this.radius = undefined;
  this.fillColours = [];
  this.vertices = [];
  this.faces = [];
  this.showAllFaces = false;
  this.halfheight = undefined;
  this.background = false;
  this.freeFall = false;
  this.velocity = new Vector3D (0, 0, 0);
  this.angular = new Vector3D (0, 0, 0);
  this.movable = false;
}

// iVerts is an array of indices into this.vertices, for the vertices enclosing face [i]
Shape.prototype.makeFace = function (i, iVerts)
{ var sumX = 0, sumY = 0, sumZ = 0, ivLen = iVerts.length;
  var faceVertices = [];
  for (var iv = 0; iv < ivLen; iv++)
  { var v = this.vertices [iVerts [iv]];
    faceVertices.push (v);
    sumX += v.x; sumY += v.y; sumZ += v.z;
  }
  var faceCentre = new Point3D (sumX / ivLen, sumY / ivLen, sumZ / ivLen); // mean values
  this.faces [i] = new Face (this, faceCentre, faceVertices, this.fillColours [i], iVerts);
};

Shape.prototype.setAngular = function (ax, ay, az) 
{ this.angular = new Vector3D (ax, ay, az); 
  this.movable = (ax !== 0 || ay !== 0 || az !== 0);
};

Shape.prototype.explode = function () 
{ for (var i = 0; i < this.faces.length; i++) { this.faces [i].explode (); } };
Shape.prototype.setColour = function (r, g, b) 
{ for (var i = 0; i < this.faces.length; i++) { this.faces [i].colour = {r:r, g:g, b:b}; } };
Shape.prototype.setStroke = function (r, g, b) 
{ for (var i = 0; i < this.faces.length; i++) { this.faces [i].stroke = {r:r, g:g, b:b}; } };
Shape.prototype.setStroking = function (bool) 
{ for (var i = 0; i < this.faces.length; i++) { this.faces [i].stroking = bool; } };
Shape.prototype.setSolid = function ()
{ if (!this.movable) for (var i = 0; i < this.faces.length; i++) { this.faces [i].setSolid (); } };
Shape.prototype.setFreeFall = function (bool) { this.freeFall = bool; };

Shape.prototype.setVelocity = function (dx, dy, dz) 
{ this.velocity = new Vector3D (dx, dy, dz); 
  this.movable = (dx !== 0 || dy !== 0 || dz !== 0);
};

Shape.prototype.view = function (me, fogNo)
{ this.minX = null; this.maxX = null; this.minY = null; this.maxY = null;
  var nFaces = this.faces.length;
  if (1 === nFaces) this.faces [0].view (me, fogNo);
  else
  { var visible = new Array (nFaces);
    for (var i = 0; i < nFaces; i++)
    { var face = this.faces [i];
      var d = face.centre.distance (me.pt);
      visible [i] = {d:d, face:face};
    }
    visible.sort (function (a, b) {return a.d - b.d;}); // Ascending distance
    if (this.showAllFaces) i = nFaces - 1;
    else i = Math.ceil (nFaces / 2); // Normally see at most half of the faces, those nearest:
    for ( ; i >= 0; i--) // Furthest first
    { visible [i].face.view (me, fogNo); }
  }
};

Shape.prototype.scale = function (factor)
{ for (var i = 0; i < this.vertices.length; i++)
  { var vec = this.vertices [i].difference (this.centre);
    this.vertices [i] = this.centre.newOffsetPt (vec.dx * factor, vec.dy * factor, vec.dz * factor);
  }
  for (i = 0; i < this.faces.length; i++) { this.faces [i].recentre (); }
  if (this.radius) this.radius *= factor;
  if (this.halfheight) this.halfheight *= factor;
};

Shape.prototype.rotateX = function (degs)
{ if (0 === degs) return;
  var rad = degs * DEG2RAD, cos = Math.cos (rad), sin = Math.sin (rad);
  for (var i = 0; i < this.vertices.length; i++)
  { var v = this.vertices [i];
    var y0 = v.y - this.centre.y, z0 = v.z - this.centre.z;
    v.y = y0 * cos + z0 * sin + this.centre.y;
    v.z = -y0 * sin + z0 * cos + this.centre.z;
  }
  for (i = 0; i < this.faces.length; i++)
  { var c = this.faces [i].centre;
    var y0 = c.y - this.centre.y, z0 = c.z - this.centre.z;
    c.y = y0 * cos + z0 * sin + this.centre.y;
    c.z = -y0 * sin + z0 * cos + this.centre.z;
    for (var j = 0; j < this.faces [i].decorations.length; j++) // New in shape6
    { var dj = this.faces [i].decorations [j];
      var cj = dj.centre;
      var y0 = cj.y - this.centre.y, z0 = cj.z - this.centre.z;
      dj.centre = new Point3D (cj.x, y0 * cos + z0 * sin + this.centre.y,
                              -y0 * sin + z0 * cos + this.centre.z);
      for (var k = 0; k < dj.vertices.length; k++)
      { var vj = dj.vertices [k];
        var y0 = vj.y - this.centre.y, z0 = vj.z - this.centre.z;
        vj.y = y0 * cos + z0 * sin + this.centre.y;
        vj.z = -y0 * sin + z0 * cos + this.centre.z;
      }
      for (k = 0; k < dj.faces.length; k++)
      { var cf = dj.faces [k].centre;
        var y0 = cf.y - this.centre.y, z0 = cf.z - this.centre.z;
        cf.y = y0 * cos + z0 * sin + this.centre.y;
        cf.z = -y0 * sin + z0 * cos + this.centre.z;
  } } }
};

Shape.prototype.rotateY = function (degs)
{ if (0 === degs) return;
  var rad = degs * DEG2RAD, cos = Math.cos (rad), sin = Math.sin (rad);
  for (var i = 0; i < this.vertices.length; i++)
  { var v = this.vertices [i];
    var x0 = v.x - this.centre.x, z0 = v.z - this.centre.z;
    v.x = x0 * cos + z0 * sin + this.centre.x;
    v.z = -x0 * sin + z0 * cos + this.centre.z;
  }
  for (i = 0; i < this.faces.length; i++)
  { var c = this.faces [i].centre;
    var x0 = c.x - this.centre.x, z0 = c.z - this.centre.z;
    c.x = x0 * cos + z0 * sin + this.centre.x;
    c.z = -x0 * sin + z0 * cos + this.centre.z;
    for (var j = 0; j < this.faces [i].decorations.length; j++) // New in shape6
    { var dj = this.faces [i].decorations [j];
      var cj = dj.centre;
      var x0 = cj.x - this.centre.x, z0 = cj.z - this.centre.z;
      dj.centre = new Point3D (x0 * cos + z0 * sin + this.centre.x, cj.y,
                              -x0 * sin + z0 * cos + this.centre.z);
      for (var k = 0; k < dj.vertices.length; k++)
      { var vj = dj.vertices [k];
        var x0 = vj.x - this.centre.x, z0 = vj.z - this.centre.z;
        vj.x = x0 * cos + z0 * sin + this.centre.x;
        vj.z = -x0 * sin + z0 * cos + this.centre.z;
      }
      for (k = 0; k < dj.faces.length; k++)
      { var cf = dj.faces [k].centre;
        var x0 = cf.x - this.centre.x, z0 = cf.z - this.centre.z;
        cf.x = x0 * cos + z0 * sin + this.centre.x;
        cf.z = -x0 * sin + z0 * cos + this.centre.z;
  } } }
};

Shape.prototype.rotateZ = function (degs)
{ if (0 === degs) return;
  var rad = degs * DEG2RAD, cos = Math.cos (rad), sin = Math.sin (rad);
  for (var i = 0; i < this.vertices.length; i++)
  { var v = this.vertices [i];
    var x0 = v.x - this.centre.x, y0 = v.y - this.centre.y;
    v.x = x0 * cos + y0 * sin + this.centre.x;
    v.y = -x0 * sin + y0 * cos + this.centre.y;
  }
  for (i = 0; i < this.faces.length; i++)
  { var c = this.faces [i].centre;
    var x0 = c.x - this.centre.x, y0 = c.y - this.centre.y;
    c.x = x0 * cos + y0 * sin + this.centre.x;
    c.y = -x0 * sin + y0 * cos + this.centre.y;
    for (var j = 0; j < this.faces [i].decorations.length; j++) // New in shape6
    { var dj = this.faces [i].decorations [j];
      var cj = dj.centre;
      var x0 = cj.x - this.centre.x, y0 = cj.y - this.centre.y;
      dj.centre = new Point3D (x0 * cos + y0 * sin + this.centre.x,
                              -x0 * sin + y0 * cos + this.centre.y, cj.z);
      for (var k = 0; k < dj.vertices.length; k++)
      { var vj = dj.vertices [k];
        var x0 = vj.x - this.centre.x, y0 = vj.y - this.centre.y;
        vj.x = x0 * cos + y0 * sin + this.centre.x;
        vj.y = -x0 * sin + y0 * cos + this.centre.y;
      }
      for (k = 0; k < dj.faces.length; k++)
      { var cf = dj.faces [k].centre;
        var x0 = cf.x - this.centre.x, y0 = cf.y - this.centre.y;
        cf.x = x0 * cos + y0 * sin + this.centre.x;
        cf.y = -x0 * sin + y0 * cos + this.centre.y;
  } } }
};

Shape.prototype.translate = function (dx, dy, dz) 
{ this.centre.offset (dx, dy, dz);
  for (var i = 0; i < this.vertices.length; i++) this.vertices [i].offset (dx, dy, dz);
  for (i = 0; i < this.faces.length; i++)
  { this.faces [i].centre.offset (dx, dy, dz);
    for (var j = 0; j < this.faces [i].decorations.length; j++)
    { this.faces [i].decorations [j].translate (dx, dy, dz);
  } }
};
