// Copyright (c) Graham Relf, UK, 2020
// www.grelf.net
'use strict';

var me, world;
/* Coordinates: x from west to east, y from south to north, z upwards from ground 0 */

function World (range, skyRGB, withFog)
{ var canvas = document.getElementById ("view");
  if (undefined === canvas) { error ('HTML page has no &lt;canvas id=&quot;view&quot;&gt;'); return; }
  canvas.width = window.innerWidth - 20;
  canvas.height = window.innerHeight - 120;
  this.g2 = canvas.getContext ("2d");
  this.screenWd = canvas.width;
  this.screenWd2 = this.screenWd / 2;
  this.screenHt = canvas.height;
  this.screenHt2 = this.screenHt / 2;
  this.canvas = canvas;
  me = new Viewer (0, 0, 8, 0); // May be overridden in run()
  this.images = {}; // To be indexed by name
  this.nImsLoaded = 0; 
  this.nImsToLoad = 0; 
  this.range = range;
  this.sky = skyRGB;
  this.cssSky = makeCssColour (skyRGB.r, skyRGB.g, skyRGB.b);
  this.withFog = withFog;
  this.g = 0.003; // gravity
  this.frameCount = 0;
  this.animIds = [];
  this.shapes = [];
  this.solidWalls = [];
/* Solid walls are faces that cannot be moved through. Currently limited to vertical walls which
 * must be aligned to x- or y-axes and must not be moveable. Shapes have method setSolid() */
  this.CLOSEST = 6; // Cannot get closer to solid walls
  this.doors = [];
  this.explosives = []; // List of shapes that could explode
  this.targets = []; // List of shapes that will respond to mouse clicks
  this.score = 0;
  this.cleanerScore = 0;
  this.animPaused = false;
  this.uiLocked = false;
}

function animate ()
{ var someMoved = world.moveShapes ();
  world.frameCount++;
  // In case there are many requests outstanding:
  // Without this the interaction becomes jerky after a while
  for (var i = 0; i < world.animIds.length; i++)
  { cancelAnimationFrame (world.animIds [i]);
    world.animIds = [];
  }
  if ((world.frameCount & 0x7ff) === 0x7ff // Every 2048
    && world.explosives.length > 0)
  { var exploder = world.explosives [0];
    world.explosives.splice (0, 1);
    exploder.explode ();
  }
  else if (someMoved && !world.animPaused) world.animIds.push (requestAnimationFrame (animate));
  world.draw ();
}

World.prototype.add = function (shape)
{ this.shapes.push (shape);
  if (shape.explosive) this.explosives.push (shape);
  if (shape.isTarget) this.targets.push (shape);
};

World.prototype.allImsLoaded = function () { return this.nImsLoaded === this.nImsToLoad; };

World.prototype.remove = function (shape)
{ for (var i = 0; i < this.shapes.length; i++)
  { if (this.shapes [i] === shape) this.shapes.splice (i, 1); }
  if (shape.explosive)
  { for (i = 0; i < this.explosives.length; i++)
    { if (this.explosives [i] === shape) this.explosives.splice (i, 1); }
  }
  shape.explosive = false;
  if (shape.isTarget)
  { for (i = 0; i < this.targets.length; i++)
    { if (this.targets [i] === shape) this.targets.splice (i, 1); }
  }
  shape.isTarget = false;
  shape.isFragment = false;
};

World.prototype.moveShapes = function ()
{ var someMove = false;
  const nShapes = this.shapes.length;
  for (var i = 0; i < nShapes; i++)
  { var shape = this.shapes [i];
if (!shape) continue;
    if (shape.movable)
    { someMove = true;
      if (shape.update) shape.update ();
      else
      { shape.rotateX (shape.angular.dx);
        shape.rotateY (shape.angular.dy);
        shape.rotateZ (shape.angular.dz);
        // Reflect off edge of world:
        if (shape.centre.x > this.range
         || shape.centre.x < -this.range) shape.velocity.dx = -shape.velocity.dx;
        if (shape.centre.y > this.range
         || shape.centre.y < -this.range) shape.velocity.dy = -shape.velocity.dy;
        if (shape.freeFall) shape.velocity.dz -= this.g;
/*        if ((this.frameCount & 0x7f) === 0) // Every 128
        { shape.velocity.rotateZ (random (-5, 5)); } // Drift */
        var newZ = shape.centre.z + shape.velocity.dz;
        if (newZ < shape.halfheight) // 0 for a Polygon
        { var frac = (shape.centre.z - shape.halfheight) / newZ;
          shape.translate (frac * shape.velocity.dx, frac * shape.velocity.dy, frac * shape.velocity.dz);
          if (shape.bouncy)
          { shape.velocity.dz = -shape.velocity.dz * 0.97;
            frac = 1 - frac;
            shape.translate (frac * shape.velocity.dx, frac * shape.velocity.dy, frac * shape.velocity.dz);
          }
          else if (this.house.over (shape.centre, 6)) this.remove (shape);
          else // fragment hits ground:
          { var removed = false;
            for (i = 0; i < shape.vertices.length; i++)
            { var svi = shape.vertices [i];
              if (svi.x <= -this.range || svi.x >= this.range
               || svi.y <= -this.range || svi.y >= this.range)
              { this.remove (shape);
                removed = true;
              }
              else
              { for (var j = 0; j < this.solidWalls.length && !removed; j++)
                { if (this.solidWalls [j].tooClose (svi))
                  { this.remove (shape);
                    removed = true;
            } } } }
            if (!removed)
            { shape.setVelocity (0, 0, 0);
              shape.setAngular (0, 0, 0);
              shape.freeFall = false;
              shape.isFragment = true;
              if (shape instanceof Polygon) shape.level ();
              shape.centre.z = 0;
              shape.isTarget = true;
              shape.response = function ()
              { world.remove (this);
                world.score += 10;
                if (world.cleaner) world.cleaner.aiming = false;
                hint ("10 points for cleaning");
              };
              this.targets.push (shape);
        } } }
        else shape.translate (shape.velocity.dx, shape.velocity.dy, shape.velocity.dz);
  } } }
  return someMove;
};

World.prototype.findShapesInRange = function ()
{ this.shapesInRange = [];
  for (var i = 0; i < this.shapes.length; i++)
  { var shape = this.shapes [i];
    var da = me.pt.distanceAndBearing (shape.centre);
    if (da.d < this.range)
    { var db = da.b - me.bRad;
      if (db < -Math.PI) db += TWO_PI; else if (db > Math.PI) db -= TWO_PI;
      if (db > -PI_BY_2 && db < PI_BY_2)
      { var fogNo = 0;
        if (this.withFog)
        { var dr = da.d / this.range;
          fogNo = (dr > 0.75) ? (dr - 0.75) * 4 : 0; // 0..1
        }
        this.shapesInRange.push ({d:da.d, fogNo:fogNo, shape:shape});
  } } }
  this.shapesInRange.sort (function (a, b) {return b.d - a.d;}); // Descending distance
};

World.prototype.draw = function ()
{ var t0 = new Date ().getTime (); // ms
  if (me.pt.z < 90) this.g2.fillStyle = this.cssSky;
  else if (me.pt.z > 150) this.g2.fillStyle = '#000';
  else
  { var ratio = (150 - me.pt.z) / 60;
    this.g2.fillStyle = makeCssColour (Math.floor (this.sky.r * ratio),
      Math.floor (this.sky.g * ratio), Math.floor (this.sky.b * ratio));
  }
  this.g2.fillRect (0, 0, this.screenWd, this.screenHt);
  this.findShapesInRange ();
  var sirL = this.shapesInRange.length;
  // Underfloor pass:
  for (var i = 0; i < sirL; i++) // Furthest first
  { var siri = this.shapesInRange [i];
    if (siri.shape.under) siri.shape.view (me, siri.fogNo);
  }
  // Background pass:
  for (i = 0; i < sirL; i++) // Furthest first
  { siri = this.shapesInRange [i];
    if (siri.shape.background) siri.shape.view (me, siri.fogNo);
  }
  // Foreground pass:
  for (i = 0; i < sirL; i++) // Furthest first
  { siri = this.shapesInRange [i];
    if (!siri.shape.background && !siri.shape.under) siri.shape.view (me, siri.fogNo);
  }
  var dt = new Date ().getTime () - t0;
  var html = me.toString () + '. ' + ' <strong>Score: ' + this.score + 
    ' Cleaner: ' + this.cleanerScore + '</strong> (Frame ' + this.frameCount + 
    ', ' + sirL + ' shapes drawn in ' + dt + 'ms)';
  document.getElementById ("info").innerHTML = html;
};

World.prototype.drawImage = function (im, x, y, wd, ht, fogNo) // fogNo 0..1
{ x |= 0; y |= 0;
  if (0 === fogNo) this.g2.drawImage (im, x, y, wd, ht); // unfogged
  else
  { var fogged, iFogNo = Math.round (fogNo * 7);
    if (undefined === im.foggy [iFogNo]) // Fogged version not yet created, create it now:
    {
      var fogged = createFoggedImageAndData (im, fogNo);
      im.foggy [iFogNo] = fogged.im;
      if (im.forFaces) im.foggy [iFogNo].data = fogged.data;
    } // Do not draw - may take time to load. It'll be missing from world for now
    else
    { fogged = im.foggy [iFogNo];
      if (fogged.loaded) this.g2.drawImage (fogged, x, y, wd, ht);
} } };
